-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;   
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all; 
use work.sha3_pkg.all;
use work.groestl_pkg.all;

-- possible generics values: hs = {GROESTL_DATA_SIZE_SMALL, GROESTL_DATA_SIZE_BIG}
-- rom_style = {DISTRIBUTED, COMBINATIONAL}
-- all combinations are allowed 

entity groestl_pq is
generic (n:integer := GROESTL_DATA_SIZE_SMALL; rom_style : integer := DISTRIBUTED);
port( 
	clk				: in std_logic;
	rst				: in std_logic;
	p_mode			: in std_logic;	
	round			: in std_logic_vector(7 downto 0);
	input 			: in std_logic_vector(n-1 downto 0);
   	output 			: out std_logic_vector(n-1 downto 0));
end groestl_pq;  	  	 	   

architecture round3_pipelined of groestl_pq is
	signal	addcons			: std_logic_vector(n-1 downto 0);
	signal	after_subbyte		: std_logic_vector(n-1 downto 0);
	signal	after_reg		: std_logic_vector(n-1 downto 0);
	signal	after_shiftrow		: std_logic_vector(n-1 downto 0);
	signal after_srp	: std_logic_vector(n-1 downto 0);
	signal after_srq	: std_logic_vector(n-1 downto 0);
	constant zero			: std_logic_vector(n-1 downto 0):=(others=>'0'); 
begin
			   
	addconsGen: 
		for i in 0 to n/64-1 generate
			addcons(i*64+63 downto i*64+ 56) <= input(i*64+63 downto i*64+56) xor round xor (conv_std_logic_vector(n/64-1-i,4) & x"0") when (p_mode = '1') else (not input(i*64+63 downto i*64+ 56));
			addcons(i*64+55 downto i*64+  8) <= input(i*64+55 downto i*64+ 8) when (p_mode = '1') else (not input(i*64+55 downto i*64+ 8));
			addcons(i*64+ 7 downto i*64+  0) <= input(i*64+ 7 downto i*64+ 0) when (p_mode = '1') else ((conv_std_logic_vector((16-n/64)+i,4) & x"F" ) xor input(i*64+ 7 downto i*64+ 0) xor round);			
		end generate;			   
	
	sbox_gen: for i in 0 to n/AES_SBOX_SIZE - 1  generate
	sbox	: aes_sbox 	generic map (rom_style=>rom_style)
			port map (	 
				input=>addcons(AES_SBOX_SIZE*i + 7 downto AES_SBOX_SIZE*i), 
				output=>after_subbyte(AES_SBOX_SIZE*i+7 downto AES_SBOX_SIZE*i));	
	end generate;	

	pl_reg		: regn generic map (n=>n, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_subbyte, output=>after_reg);

	srp			:entity work.groestl_shiftrow(groestl_shiftrowp) generic map (n=>n)port map (input=>after_reg, output=>after_srp);
	srq			:entity work.groestl_shiftrow(groestl_shiftrowq) generic map (n=>n)port map (input=>after_reg, output=>after_srq);
	
	after_shiftrow <= after_srp when p_mode='0' else after_srq;

	mc:
		entity work.groestl_mixbytes(groestl_mixbytes)	
		generic map (n=>n)
		port map (input=>after_shiftrow,  output=>output);	

end round3_pipelined; 			 	   
